import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:rxdart/rxdart.dart';

abstract class AgoraAction {
  final String channelId;

  AgoraAction(this.channelId);
}

class AgoraJoinAction extends AgoraAction {
  final int userId;
  AgoraJoinAction(String channelId, this.userId) : super(channelId);
}

class AgoraJoinedAction extends AgoraAction {
  final int userId;

  AgoraJoinedAction(String channeld, this.userId) : super(channeld);
}

class AgoraLeftAction extends AgoraAction {
  final String erorr;
  AgoraLeftAction(String channelId, this.erorr) : super(channelId);
}

class AgoraFriendJoinedAction extends AgoraAction {
  final int friendId;

  AgoraFriendJoinedAction(String channelId, this.friendId) : super(channelId);
}

class AgoraFriendLeftAction extends AgoraAction {
  final int friendId;

  AgoraFriendLeftAction(String channelId, this.friendId) : super(channelId);
}

class AgoraAudioMutedAction extends AgoraAction {
  final int friendId;
  final bool mute;
  final bool isLocal;
  AgoraAudioMutedAction(
      String channelId, this.friendId, this.mute, this.isLocal)
      : super(channelId);
}

class AgoraVideoMutedAction extends AgoraAction {
  final int friendId;
  final bool mute;
  final bool isLocal;
  AgoraVideoMutedAction(
      String channelId, this.friendId, this.mute, this.isLocal)
      : super(channelId);
}

class AgoraSpeakerMutedAction extends AgoraAction {
  final bool muted;
  AgoraSpeakerMutedAction(
    String channelId,
    this.muted,
  ) : super(channelId);
}

class AgoraSpeakerChangeAction extends AgoraAction {
  final int userId;

  AgoraSpeakerChangeAction(String channelId, this.userId) : super(channelId);
}

class AgoraTimerEvent extends AgoraAction {
  final int timerInSeconds;

  AgoraTimerEvent(String channelId, this.timerInSeconds) : super(channelId);
}

///
/// Agora State
///

abstract class AgoraState {
  final String channelId;

  AgoraState(this.channelId);
}

class AgoraIdleState extends AgoraState {
  AgoraIdleState() : super('');
}

class AgoraCurrentState extends AgoraState {
  final int currentId;
  final Set<int> members;
  final Set<int> audioMuted;
  final Set<int> videoMuted;
  final int activeSpeaker;
  final bool isSpeakerOn;
  final int timerInSeconds;

  AgoraCurrentState(
      String channelId,
      this.currentId,
      this.members,
      this.audioMuted,
      this.videoMuted,
      this.activeSpeaker,
      this.isSpeakerOn,
      this.timerInSeconds)
      : super(channelId);

  AgoraCurrentState copyWith(
      {String channelId,
      int currentId,
      Set<int> members,
      Set<int> audioMuted,
      Set<int> videoMuted,
      int activeSpeaker,
      bool isSpeakerOn,
      int timerInSeconds}) {
    return AgoraCurrentState(
        channelId ?? this.channelId,
        currentId ?? this.currentId,
        members ?? this.members,
        audioMuted ?? this.audioMuted,
        videoMuted ?? this.videoMuted,
        activeSpeaker ?? this.activeSpeaker,
        isSpeakerOn ?? this.isSpeakerOn,
        timerInSeconds ?? this.timerInSeconds);
  }

  bool get isAudioMuted => audioMuted.contains(currentId);

  bool get isVideoMuted => videoMuted.contains(currentId);

  static AgoraCurrentState get emptyState {
    return AgoraCurrentState(
        null, null, Set<int>(), Set<int>(), Set<int>(), null, false, 0);
  }
}

class AgoraEndState extends AgoraState {
  AgoraEndState(String channelId) : super(channelId);
}

///
/// Bloc
///

class AgoraBloc extends Bloc<AgoraAction, AgoraState> {
  AgoraBloc() : super(AgoraIdleState());

  @override
  Stream<AgoraState> mapEventToState(AgoraAction event) async* {
    if (event is AgoraJoinAction) {
      yield AgoraCurrentState.emptyState;
    }
    if (!(state is AgoraCurrentState)) {
      return;
    }
    if (event is AgoraLeftAction) {
      yield AgoraIdleState();
      return;
    }
    AgoraCurrentState currentState = state;
    if (event is AgoraJoinedAction) {
      Set<int> members = currentState.members ?? Set<int>();
      members.add(event.userId);
      yield currentState.copyWith(
          members: members,
          currentId: event.userId,
          channelId: event.channelId);
    }
    if (event is AgoraAudioMutedAction) {
      AgoraCurrentState currentState = state as AgoraCurrentState;
      Set<int> muted = Set<int>.from(currentState.audioMuted);
      if (event.mute) {
        muted.add(event.friendId);
      } else {
        muted.remove(event.friendId);
      }
      yield currentState.copyWith(audioMuted: muted);
    }
    if (event is AgoraFriendJoinedAction) {
      Set<int> members = Set<int>.from(currentState.members ?? Set<int>());
      if (!members.contains(event.friendId)) {
        members.add(event.friendId);
      }
      yield currentState.copyWith(members: members);
    }
    if (event is AgoraFriendLeftAction) {
      Set<int> muted = Set<int>.from(currentState.videoMuted);
      if (muted.contains(event.friendId)) {
        muted.remove(event.friendId);
      }
      Set<int> audioMuted = Set<int>.from(currentState.audioMuted);
      if (audioMuted.contains(event.friendId)) {
        audioMuted.remove(event.friendId);
      }
      Set<int> members = Set<int>.from(currentState.members ?? Set<int>());
      if (members.contains(event.friendId)) {
        members.remove(event.friendId);
      }
      yield currentState.copyWith(
          videoMuted: muted, audioMuted: audioMuted, members: members);
    }

    if (event is AgoraVideoMutedAction) {
      Set<int> muted = Set<int>.from(currentState.videoMuted);
      if (event.mute) {
        muted.add(event.friendId);
      } else {
        muted.remove(event.friendId);
      }
      yield currentState.copyWith(videoMuted: muted);
    }
    if (event is AgoraSpeakerMutedAction) {
      yield currentState.copyWith(isSpeakerOn: event.muted);
    }
    if (event is AgoraSpeakerChangeAction) {
      yield currentState.copyWith(activeSpeaker: event.userId);
    }
    if (event is AgoraTimerEvent) {
      yield currentState.copyWith(timerInSeconds: event.timerInSeconds);
    }
    debugPrint("AGORA EVENT - $event ");
  }

  @override
  void onError(Object error, StackTrace stackTrace) {
    super.onError(error, stackTrace);
  }

  @override
  void onEvent(AgoraAction event) {
    AgoraEventStream.addAction(event);
    super.onEvent(event);
  }
}

class AgoraEventStream {
  static BehaviorSubject<AgoraAction> _eventSubject;

  static void open() {
    if (_eventSubject == null) _eventSubject = BehaviorSubject<AgoraAction>();
  }

  static void addAction(AgoraAction action) {
    _eventSubject.add(action);
  }

  static Stream<AgoraAction> get stream => _eventSubject.stream;

  static void close() {
    if (_eventSubject != null) {
      if (!_eventSubject.isClosed) _eventSubject.close();
      _eventSubject = null;
    }
  }
}
